import logging
import os
from typing import Union, Tuple, Dict

import numpy as np
import pygame

from gui import TextLayer
from gui.opengl import OpenGLMesh, OpenGLGUIFacade


class OpenGLTextLayer(OpenGLMesh, TextLayer):

    def __init__(self, gui: OpenGLGUIFacade):
        super(OpenGLTextLayer, self).__init__(gui)
        self.__vertices = None  # type: Union[None, np.ndarray]
        self.__faces = None  # type: Union[None, np.ndarray]
        self.__uvMap = None  # type: Union[None, np.ndarray]
        self.__transparentTile = (0, 0)  # type: Tuple[int, int]
        self.__characterTiles = {}  # type: Dict[str, (int, int)]

    @property
    def maxCharacterCount(self) -> int:
        return self.__vertices.shape[0]

    def setFont(self, fontName: str, tileHeight: int, color: (int, int, int)):
        characters = [
            " !\"#$%&'()*+,-./",
            "0123456789:;<=>?",
            "@ABCDEFGHIJKLMNO",
            "PQRSTUVWXYZ[\\]^_",
            "`abcdefghijklmno",
            "pqrstuvwxyz{|}~ ",
        ]
        charactersPerRow = 16
        rowCount = len(characters)

        # Render a single character to find the required tile height
        fontSize = int(tileHeight * 1.3)
        for iterations in range(100):  # Up to 100 tries
            font = pygame.font.Font(fontName, fontSize)
            surface = font.render("M", False, color)
            if surface.get_height() == tileHeight:
                break
            elif surface.get_height() > tileHeight:
                fontSize -= 1
            else:
                fontSize += 1
        tileWidth = surface.get_width()
        tileHeight = surface.get_height()

        # Create a square image with width and height equal to a power of two
        def nextPowerOfTwo(x: int) -> int:
            return 1 if x == 0 else 2 ** (x - 1).bit_length()
        imageWidth = nextPowerOfTwo(charactersPerRow * tileWidth)
        imageHeight = nextPowerOfTwo(rowCount * tileHeight)
        imageSize = max(imageWidth, imageHeight)
        image = pygame.Surface((imageSize, imageSize), flags=pygame.SRCALPHA)

        self.__characterTiles = {}
        for rowIndex in range(rowCount):
            for colIndex, character in enumerate(characters[rowIndex]):
                # Render and blit the character
                surface = font.render(character, False, color)
                if surface.get_width() > tileWidth:
                    logging.warning("Character {} larger width {} > {}".format(character, surface.get_width(), tileWidth))
                image.blit(surface, (colIndex * tileWidth, rowIndex * tileHeight))
                # Compute the tile location for each character in the row
                self.__characterTiles[character] = (colIndex, rowIndex)

        # Set transparent tile on space
        self.__transparentTile = self.__characterTiles[" "]

        # Convert to numpy array
        imageArray = np.zeros((imageSize, imageSize, 4), dtype=np.uint8)
        imageArray[..., 0] = pygame.surfarray.pixels_red(image)
        imageArray[..., 1] = pygame.surfarray.pixels_green(image)
        imageArray[..., 2] = pygame.surfarray.pixels_blue(image)
        imageArray[..., 3] = pygame.surfarray.pixels_alpha(image)
        imageArray = imageArray.transpose((1, 0, 2))

        # Send the tileset to the GPU
        self.setTileset(imageArray, tileWidth, tileHeight)

        if False: # Save generated tileset (to debug/illustrate)
            from PIL import Image
            Image.fromarray(imageArray).save("{}-{}.png".format(os.path.splitext(fontName)[0], tileHeight))

    def setMaxCharacterCount(self, maxCount: int):
        if self.__vertices is not None:
            raise RuntimeError("Update of max character count is not supported")

        self.__vertices = np.empty([maxCount, 4, 2], dtype=np.float32)
        self.__uvMap = np.empty([maxCount, 4, 2], dtype=np.float32)
        self.__faces = np.empty([maxCount, 4], dtype=np.uint)
        faceCount = 0
        for quadIndex in range(maxCount):
            self.__setQuadLocation(quadIndex, 0, 0)
            self.__setQuadTile(quadIndex, self.__transparentTile[0], self.__transparentTile[1])

            self.__faces[quadIndex, 0] = faceCount * 4
            self.__faces[quadIndex, 1] = faceCount * 4 + 1
            self.__faces[quadIndex, 2] = faceCount * 4 + 2
            self.__faces[quadIndex, 3] = faceCount * 4 + 3
            faceCount += 1

        self.setData(self.__vertices, self.__faces, self.__uvMap, dynamic=True)

    def draw(self):
        self.updateData(self.__vertices, self.__faces, self.__uvMap)
        super(OpenGLTextLayer, self).draw()

    def setText(self, x: float, y: float, content: str):
        if len(content) > self.maxCharacterCount:
            content = content[0:self.maxCharacterCount]

        x0 = x
        # Characters to display
        for charIndex, character in enumerate(content):
            if character == "\n":
                x = x0
                y += self.tileHeight
                continue
            if character in self.__characterTiles:
                tileCoords = self.__characterTiles[character]
            else:
                tileCoords = self.__characterTiles["?"]
            self.__setQuadTile(charIndex, tileCoords[0], tileCoords[1])
            self.__setQuadLocation(charIndex, x, y)
            x += self.tileWidth

        # Remaining characters are transparent
        for charIndex in range(len(content), self.maxCharacterCount):
            self.__setQuadTile(charIndex, self.__transparentTile[0], self.__transparentTile[1])

    def __setQuadLocation(self, quadIndex: int, x: float, y: float):
        assert 0 <= quadIndex < self.__vertices.shape[0]

        spriteScreenX1 = -1 + x * self.screenPixelWidth
        spriteScreenY1 = 1 - y * self.screenPixelHeight
        spriteScreenX2 = spriteScreenX1 + self.screenTileWidth
        spriteScreenY2 = spriteScreenY1 - self.screenTileHeight
        self.__vertices[quadIndex, 0] = [spriteScreenX1, spriteScreenY2]
        self.__vertices[quadIndex, 1] = [spriteScreenX1, spriteScreenY1]
        self.__vertices[quadIndex, 2] = [spriteScreenX2, spriteScreenY1]
        self.__vertices[quadIndex, 3] = [spriteScreenX2, spriteScreenY2]

    def __setQuadTile(self, quadIndex: int, tileX: float, tileY: float):
        assert 0 <= quadIndex < self.__vertices.shape[0]

        spriteTextureX1 = tileX * self.textureTileWidth
        spriteTextureY1 = tileY * self.textureTileHeight
        spriteTextureX2 = spriteTextureX1 + self.textureTileWidth
        spriteTextureY2 = spriteTextureY1 + self.textureTileHeight
        self.__uvMap[quadIndex, 0] = [spriteTextureX1, spriteTextureY2]
        self.__uvMap[quadIndex, 1] = [spriteTextureX1, spriteTextureY1]
        self.__uvMap[quadIndex, 2] = [spriteTextureX2, spriteTextureY1]
        self.__uvMap[quadIndex, 3] = [spriteTextureX2, spriteTextureY2]
